﻿using System;
using System.Linq;
using kidsvt_lab7_8.Enums;

namespace kidsvt_lab7_8
{
    class MemoryUnit
    {
        public int NumberOfRows { get; private set; }
        public int NumberOfColumns { get; private set; }

        private readonly bool[][] _memoryArray;

        private readonly FaultType _faultType;

        private MemoryCell[] _faultyCellsArray;      // Ячейки с неисправностями
        private int[] _faultyCellsInfoArray;
        private Random _random;

        public int FaultsCount { get; private set; }
        public int AddressFaultsCount { get; private set; }

        public MemoryUnit(int numberOfRows, int numberOfColumns, FaultType faultType, int faultsCount)
        {
            NumberOfRows = numberOfRows;
            NumberOfColumns = numberOfColumns;
            _faultType = faultType;
            FaultsCount = faultsCount;
            AddressFaultsCount = NumberOfRows/2*NumberOfColumns/2;

            _memoryArray = new bool[numberOfRows][];
            for (int i = 0; i < numberOfRows; i++)
                _memoryArray[i] = new bool[numberOfColumns];

            GenerateFaultyCells();
        }

        private void GenerateFaultyCells()
        {
            _faultyCellsArray = new MemoryCell[FaultsCount];
            _random = new Random();

            for (int i = 0; i < FaultsCount; i++)
            {
                MemoryCell cell;

                do cell = GetRandomMemoryCell();
                while (_faultyCellsArray.Contains(cell));

                _faultyCellsArray[i] = cell;
            }

            if (_faultType == FaultType.SAF)
            {
                _faultyCellsInfoArray = new int[FaultsCount];
                for (int i = 0; i < FaultsCount; i++)
                {
                    bool value;
                    if (_random.NextDouble() < 0.5)
                        value = false;
                    else
                        value = true;

                    _faultyCellsInfoArray[i] = Convert.ToInt32(value);
                    _memoryArray[_faultyCellsArray[i].Row][_faultyCellsArray[i].Column] = value;
                }
            }
            else if (_faultType == FaultType.PNPSFK3)
            {
                _faultyCellsInfoArray = new int[FaultsCount];

                // Трёхбитное число: старший бит - вертикальные или горизонтальные соседи,
                //                   младшие два бита - код соседей, при котором базовая ячейка не может быть изменена
                for (int i = 0; i < FaultsCount; i++)
                    _faultyCellsInfoArray[i] = _random.Next(0, 7);
            }
        }


        private MemoryCell GetRandomMemoryCell()
        {
            int row = _random.Next(1, NumberOfRows - 2);
            int column = _random.Next(1, NumberOfColumns - 2);

            return new MemoryCell(row, column);
        }

        public bool GetMemoryCellValue(int row, int column)
        {
            return _memoryArray[row][column];
        }


        public void SetMemoryCellValue(int row, int column, bool value)
        {
            MemoryCell cell = new MemoryCell(row, column);

            if (_faultType == FaultType.AF)         // Address Decoder Faults
            {
                _memoryArray[row][column] = value;
                if (row%2 != 0 || column%2 != 0) return;

                if(row + 1 < NumberOfRows && column + 1 < NumberOfColumns)
                    _memoryArray[row + 1][column + 1] = value;
            }

            else if (_faultType == FaultType.SAF)   // Stuck-At Faults
            {
                int index = Array.IndexOf(_faultyCellsArray, cell);
                if (index == -1)    // Ячейка исправна
                    _memoryArray[row][column] = value;
            }

            else if (_faultType == FaultType.CFin)  // Inversion Coupling Faults
            {
                int index = Array.IndexOf(_faultyCellsArray, cell);
                if (index != -1)
                {
                    bool firstValue = (row % 2 == 0);
                    bool secondValue = !firstValue;

                    if (_memoryArray[row][column] == firstValue && value == secondValue)
                    {
                        if (index + 1 >= _faultyCellsArray.Length)
                            index = 0;

                        int victimRow = _faultyCellsArray[index + 1].Row;
                        int victimColumn = _faultyCellsArray[index + 1].Column;

                        _memoryArray[victimRow][victimColumn] = !_memoryArray[victimRow][victimColumn];
                    }
                }
                _memoryArray[row][column] = value;
            }

            else if (_faultType == FaultType.CFid)  // Idempotent Coupling Faults
            {
                int index = Array.IndexOf(_faultyCellsArray, cell);
                if (index != -1)
                {
                    bool firstValue = (row % 2 == 0);
                    bool secondValue = !firstValue;

                    if (_memoryArray[row][column] == firstValue && value == secondValue)
                    {
                        if (index + 1 >= _faultyCellsArray.Length)
                            index = 0;

                        int victimRow = _faultyCellsArray[index + 1].Row;
                        int victimColumn = _faultyCellsArray[index + 1].Column;

                        _memoryArray[victimRow][victimColumn] = (column % 2 == 0);
                    }
                }
                _memoryArray[row][column] = value;
            }

            else if (_faultType == FaultType.PNPSFK3)   // Passive NPSF
            {
                int index = Array.IndexOf(_faultyCellsArray, cell);
                if (index != -1)
                {
                    int code = _faultyCellsInfoArray[index];
                    bool isHorizontally = ((code & 4) == 0);
                    bool firstValue = ((code & 2) == 2);
                    bool secondValue = ((code & 1) == 1);

                    int neighborRow1, neighborColumn1, neighborRow2, neighborColumn2;
                    if (isHorizontally)
                    {
                        neighborRow1 = neighborRow2 = _faultyCellsArray[index].Row;
                        neighborColumn1 = _faultyCellsArray[index].Column - 1;
                        neighborColumn2 = _faultyCellsArray[index].Column + 1;
                        if (neighborColumn1 >= 0 && neighborColumn2 < NumberOfColumns)
                        {
                            if(_memoryArray[neighborRow1][neighborColumn1] == firstValue &&
                                _memoryArray[neighborRow2][neighborColumn2] == secondValue)
                            return;
                        }
                    }
                    else
                    {
                        neighborRow1 = _faultyCellsArray[index].Row - 1;
                        neighborRow2 = _faultyCellsArray[index].Row + 1;
                        neighborColumn1 = neighborColumn2 = _faultyCellsArray[index].Column;
                        if (neighborRow1 >= 0 && neighborRow2 < NumberOfRows)
                        {
                            if(_memoryArray[neighborRow1][neighborColumn1] == firstValue &&
                                _memoryArray[neighborRow2][neighborColumn2] == secondValue)
                            return;
                        }
                    }
                }

                _memoryArray[row][column] = value;
            }

            else if (_faultType == FaultType.ANPSFK5) // Active NPSF
            {
                if (_memoryArray[row][column] == value)
                    return;

                int baseRow = row, baseColumn = column + 1;
                MemoryCell baseCell = new MemoryCell(baseRow, baseColumn);
                if (_faultyCellsArray.Contains(baseCell))
                    _memoryArray[baseRow][baseColumn] = !_memoryArray[baseRow][baseColumn];

                baseRow = row;
                baseColumn = column - 1;
                baseCell = new MemoryCell(baseRow, baseColumn);
                if (_faultyCellsArray.Contains(baseCell))
                    _memoryArray[baseRow][baseColumn] = !_memoryArray[baseRow][baseColumn];

                baseRow = row - 1;
                baseColumn = column;
                baseCell = new MemoryCell(baseRow, baseColumn);
                if (_faultyCellsArray.Contains(baseCell))
                    _memoryArray[baseRow][baseColumn] = !_memoryArray[baseRow][baseColumn];

                baseRow = row + 1;
                baseColumn = column;
                baseCell = new MemoryCell(baseRow, baseColumn);
                if (_faultyCellsArray.Contains(baseCell))
                    _memoryArray[baseRow][baseColumn] = !_memoryArray[baseRow][baseColumn];


                _memoryArray[row][column] = value;
            }
        }
    }
}
